// Beispiel zu Kapitel 16 in "Go To Delphi 5", A.Bohne / G.Lang
//  Addison-Wesley-Longman 1999

// Beispiel zur Verwendung von Sockets
// (Damespiel aus Kapitel 6 ber Netzwerk)
// HINWEIS: Das Damespiel ist nicht vollstndig implementiert.
// Es fehlt insbesondere die Verwandlung von Spielsteinen
// in Damen und eine Behandlung des Spielendes

unit CheckerUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, Menus, ScktComp;

const
  NCells = 8;

type
  TCellState = (csNone, csEmpty, csBlack, csWhite);
  TMoveRecord = record
    fromCell: TPoint;
    toCell: TPoint;
  end;
  TCheckerForm = class(TForm)
    CheckerDrawGrid: TDrawGrid;
    MainMenu1: TMainMenu;
    Spiel1: TMenuItem;
    NewMI: TMenuItem;
    N1: TMenuItem;
    ExitMI: TMenuItem;
    ClientSocket: TClientSocket;
    ServerSocket: TServerSocket;
    ConnectMI: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure CheckerDrawGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormDestroy(Sender: TObject);
    procedure CheckerDrawGridDragOver(Sender, Source: TObject; X,
      Y: Integer; State: TDragState; var Accept: Boolean);
    procedure CheckerDrawGridDragDrop(Sender, Source: TObject; X,
      Y: Integer);
    procedure CheckerDrawGridMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure CheckerDrawGridEndDrag(Sender, Target: TObject; X,
      Y: Integer);
    procedure ExitMIClick(Sender: TObject);
    procedure NewMIClick(Sender: TObject);
    procedure ConnectMIClick(Sender: TObject);
    procedure SocketRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure ServerSocketClientConnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ServerSocketClientDisconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ClientSocketDisconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ClientSocketConnect(Sender: TObject;
      Socket: TCustomWinSocket);
  private
    { Private-Deklarationen }
    CellWidth: Integer;
    CellOffset: Integer;
    CellState: array[0..NCells - 1, 0..NCells - 1] of TCellState;
    Palette: HPalette;
    DragRow, DragCol: Integer;
    LastRow, LastCol: Integer;
    WhiteMoves: Boolean;
    IsDragging: Boolean;
    Playing: Boolean;
    Started: Boolean;
    Eaten: Boolean;
    NrPieces: array[False..True] of Integer;
    Servername: string;
    Connected: Boolean;
    IsServer: Boolean;
    ShallMove: Boolean;
    procedure DrawChecker(Rect: TRect; white: Boolean);
    procedure InitPalette;
    procedure InitBoard;
    function CanMove(iCol, iRow: Integer): Boolean;
    function BeginMove(iCol, iRow: Integer): Boolean;
    procedure Move(FromCell, ToCell: TPoint);
    procedure EndMove(iCol, iRow: Integer);
    procedure UpdateCaption;
    function CanEatNext(iCol, iRow: Integer): Boolean;
  public
    { Public-Deklarationen }
  end;

var
  CheckerForm: TCheckerForm;

implementation

{$R *.DFM}
{$R CURSOR.RES}

const
  NColors = 16;
  crChecker = 1;

{$R-}

procedure TCheckerForm.InitPalette;
var
  i: Integer;
  logPalette: PLogPalette;
begin
  // Speicher fr Record anfordern
  GetMem(logPalette, SizeOf(TLogPalette) +
                     SizeOf(TPaletteEntry) * (NColors - 1));
  with logPalette^ do
  begin
    palVersion := $300;         // Versionsnummer
    palNumEntries := NColors;   // Anzahl Farben
    // Farben fr einzelne Eintrge
    for i := 0 to NColors - 1 do with palPalEntry[i] do
    begin
      peRed := i shl 4;
      peGreen := peRed;
      peBlue := peRed;
      peFlags := 0;
    end;
  end;
  // logische Palette erstellen
  Palette := CreatePalette(LogPalette^);
  FreeMem(logPalette, SizeOf(TLogPalette) +
                      SizeOf(TPaletteEntry) * (NColors - 1));
end;

{$R+}

procedure TCheckerForm.InitBoard;
// Initialisierung der Felder zu Spielbeginn
var
 row, col: Integer;
begin
  if Playing and Started then
  begin
    if MessageDlg('Spiel abbrechen?', mtConfirmation,
                   [mbYes, mbNo], 0) <> mrYes then Exit;
    Playing := False;
  end;
  for row := 0 to NCells - 1 do
    for col := 0 to NCells - 1 do
      if Odd(row + col) then CellState[col, row] := csEmpty
      else CellState[col, row] := csNone;
  for row := 0 to 2 do
    for col := 0 to NCells - 1 do
      if Odd(row + col) then CellState[col, row] := csWhite;
  for row := NCells - 3 to NCells - 1 do
    for col := 0 to NCells - 1 do
      if Odd(row + col) then CellState[col, row] := csBlack;
  CheckerDrawGrid.Invalidate;
  Started := False;
  WhiteMoves := True;
  Eaten := False;
  IsDragging := False;
  NrPieces[False] := 12;
  NrPieces[True] := 12;
end;

procedure TCheckerForm.FormCreate(Sender: TObject);
begin
  // Brettgre einstellen
  with CheckerDrawGrid do
  begin
    DefaultColWidth := Screen.Height div 16;
    DefaultRowHeight := DefaultColWidth;
    Left := 0; Top := 0;
    Width := ColCount * DefaultColWidth + GetSystemMetrics(SM_CXEDGE) * 2;
    Height := RowCount * DefaultRowHeight + GetSystemMetrics(SM_CYEDGE) * 2;
  end;
  ClientWidth := CheckerDrawGrid.Width;
  ClientHeight := CheckerDrawGrid.Height;
  CellWidth := CheckerDrawGrid.DefaultColWidth;
  CellOffset := CellWidth div 16;
  InitPalette;
  Playing := True;
  Started := False;
  InitBoard;
  Screen.Cursors[crChecker] := LoadCursor(HInstance, 'DRAGCURSOR');
  CheckerDrawGrid.DragCursor := crChecker;
  ServerSocket.Active := True;
  Connected := False;
  IsServer := True;
  ShallMove := False;
  UpdateCaption;
end;

procedure TCheckerForm.FormDestroy(Sender: TObject);
begin
  ServerSocket.Close;
  ClientSocket.Close;
  DeleteObject(Palette);
end;

procedure TCheckerForm.CheckerDrawGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
begin
  with CheckerDrawGrid.Canvas do
  begin
    // Hintergrund fllen
    if Odd(ARow + ACol) then Brush.Color := clMaroon
    else Brush.Color := clYellow;
    FillRect(Rect);
    // Stein malen
    if CellState[ACol, ARow] in [csBlack, csWhite] then
    begin
      InflateRect(rect, -CellOffset, -CellOffset);
      DrawChecker(Rect, CellState[ACol, ARow] = csWhite);
    end;
  end;
end;

procedure TCheckerForm.DrawChecker(Rect: TRect; white: Boolean);
const
  NCircles = 18;
  Colors: array [0..NCircles - 1] of TColor =
    (0, 6, 5, 4, 3, 2, 1, 0, 4, 7, 4, 0, 2, 4, 6, 2, 1, 0);
var
  i: Integer;
  w: Integer;
  wrect: Integer;
  oldPalette: HPalette;
begin
  wrect := Rect.Right - Rect.Left;
  with CheckerDrawGrid.Canvas do
  begin
    oldPalette := SelectPalette(Handle, Palette, False);
    RealizePalette(Handle);
    for i := 0 to NCircles - 1 do
    begin
      w := Round(wrect * (1.0 - 0.025 * i));
      if white then Brush.Color := PaletteIndex(Colors[i] + 8)
      else Brush.Color := PaletteIndex(Colors[i]);
      Pen.Color := Brush.Color;
      Ellipse(Rect.Left + w, Rect.Top + w,
              Rect.Right - w, Rect.Bottom - w);
    end;
    SelectPalette(Handle, oldPalette, False);
  end;
end;

procedure TCheckerForm.CheckerDrawGridDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
var
  iCol, iRow: Integer;
begin
  if Source is TDrawGrid then
  begin
    CheckerDrawGrid.MouseToCell(X, Y, iCol, iRow);
    Accept := (iCol >= 0) and (iCol < NCells)
               and (iRow >= 0) and (iRow < NCells)
               and CanMove(iCol, iRow);
  end
  else
    Accept := False;
end;

function TCheckerForm.CanEatNext(iCol, iRow: Integer): Boolean;
// nachschauen, ob nach dem Schalgen ein weiterer Stein
// geschlagen werden kann
var
  col, row, offset: Integer;
begin
  Result := False;
  if not eaten then Exit;

  if WhiteMoves then offset := 1 else offset := -1;

  // Schlagmglichkeit nach schrg links berprfen
  // (auf Stein des Gegners und leeres Feld dahinter prfen)
  row := iRow + offset;
  col := iCol - 1;
  if (col > 0) and (row > 0) and (row < NCells - 1)
    and (CellState[col, row] in [csBlack, csWhite])
    and ((CellState[col, row] = csBlack) = WhiteMoves)
    and (CellState[col - 1, row + offset] = csEmpty)
    then Result := True;

  // das gleiche fr die Schlagmglchkeit nach schrg rechts
  col := iCol + 1;
  if (col < NCells - 1) and (row > 0) and (row < NCells - 1)
    and (CellState[col, row] in [csBlack, csWhite])
    and ((CellState[col, row] = csBlack) = WhiteMoves)
    and (CellState[col + 1, row + offset] = csEmpty)
    then Result := True;

end;

procedure TCheckerForm.EndMove(iCol, iRow: Integer);
var
  rect: TRect;
  msg: string;
begin
  // Zielfeld bekommt Status von Ausgangsfeld
  CellState[iCol, iRow] := CellState[DragCol, DragRow];
  LastRow := iRow; LastCol := iCol;
  // Ausgangsfeld ist jetzt leer
  CellState[DragCol, DragRow] := csEmpty;
  with CheckerDrawGrid do
  begin
    // Zielfeld neu zeichnen lassen
    rect := CellRect(iCol, iRow);
    InvalidateRect(Handle, @rect, True);
    // Schlagen behandeln
    if Abs(iRow - DragRow) = 2 then
    begin
      if Eaten and (WhiteMoves <>
          (CellState[iCol, iRow] = csWhite)) then
        WhiteMoves := not WhiteMoves;
      iCol := (iCol + DragCol) div 2;
      iRow := (iRow + DragRow) div 2;
      // Feld mit geschlagener Figur leeren
      CellState[iCol, iRow] := csEmpty;
      rect := CellRect(iCol, iRow);
      InvalidateRect(Handle, @rect, True);
      Eaten := True;
      Dec(NrPieces[not WhiteMoves]);
      // alle Steine einer Farbe geschlagen
      if NrPieces[not WhiteMoves] = 0 then
      begin
        if IsServer = WhiteMoves then msg := 'gewonnen'
        else msg := 'leider verloren';
        ShowMessage('Sie haben ' + msg + '!');
        Playing := False;
        Exit;
      end;
    end
    else Eaten := False;
  end;
  // Wenn nach dem Schlagen eine weitere Schlagmglichkeit
  // besteht, so bleibt die Kontrolle bei det schlagenden
  // Seite - sie wird also zum weiteren Schlagen gezwungen!
  if not CanEatNext(lastCol, LastRow) then
    WhiteMoves := not WhiteMoves;
  Started := True;
end;

procedure TCheckerForm.CheckerDrawGridDragDrop(Sender, Source: TObject; X,
  Y: Integer);
// Nach Fallenlassen Spielfeld aktualisieren
var
  iCol, iRow: Integer;
  buf: TMoveRecord;
begin
  CheckerDrawGrid.MouseToCell(X, Y, iCol, iRow);
  EndMove(iCol, iRow);
  IsDragging := False;
  with buf do begin
    fromCell.x := DragCol; fromCell.y := DragRow;
    toCell.x := iCol;      toCell.y := iRow;
  end;
  if Connected then begin
    if IsServer then
      ServerSocket.Socket.Connections[0].SendBuf(buf, SizeOf(buf))
    else
      ClientSocket.Socket.SendBuf(buf, SizeOf(buf));
    ShallMove := not (IsServer xor WhiteMoves); //False;
    UpdateCaption;
  end;
  Started := True;
end;

function TCheckerForm.BeginMove(iCol, iRow: Integer):Boolean;
// beginnt Bewegung, wenn erlaubt
var
  state: TCellState;
  white: boolean;
  rect: TRect;
begin
  state := CellState[DragCol, DragRow];
  white := state = csWhite;
  // Der geklammerte Code hatte es ermglicht, da nach dem
  // Schlagen bei Bedarf ein weiteres Schlagen erfolgen
  // konnte - das geht im Netzwerkmodus so nicht, da immer
  // klar sein mu, wer am Zug ist
  if (WhiteMoves = white) {or Eaten and (DragCol = LastCol)
        and (DragRow = LastRow)} then
  with CheckerDrawGrid do begin
      CellState[DragCol, DragRow] := csEmpty;
      rect := CellRect(DragCol, DragRow);
      InvalidateRect(Handle, @rect, False);
      Update;
      CellState[DragCol, DragRow] := state;
      Result := True;
  end
  else Result := False;
end;

procedure TCheckerForm.CheckerDrawGridMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
// beginnt Ziehen, wenn erlaubt
begin
  if Playing and ShallMove and (Button = mbLeft)
  then with CheckerDrawGrid do
  begin
    MouseToCell(X, Y, DragCol, DragRow);
    if BeginMove(DragCol, DragRow) then
    begin
      IsDragging := True;
      BeginDrag(True);
    end;
  end;
end;

procedure TCheckerForm.CheckerDrawGridEndDrag(Sender, Target: TObject; X,
  Y: Integer);
// zeichnet bei Abbruch gezogenen Stein an Ausgangsposition
var
  rect: TRect;
begin
  if IsDragging then with CheckerDrawGrid do
  begin
    rect := CellRect(DragCol, DragRow);
    InvalidateRect(Handle, @rect, False);
    IsDragging := False;
  end;
end;

function TCheckerForm.CanMove(iCol, iRow: Integer): Boolean;
// prft, ob ein Zug erlaubt ist
var
  white: Boolean;
  state: TCellState;
begin
  Result := False;
  // nur auf leere Zellen ziehen lassen
  if CellState[iCol, iRow] <> csEmpty then Exit;
  white := CellState[DragCol, DragRow] = csWhite;
  // Rckwrtsgehen verbieten
  if white and (iRow <= DragRow) or
    not white and (iRow >= DragRow) then Exit;
  case Abs(iRow - DragRow) of
    // einfacher Zug
    1: if (white <> WhiteMoves) or (Abs(iCol - DragCol) <> 1) then Exit;
    // Schlagen
    2:
    begin
      if Abs(iCol - DragCol) <> 2 then Exit;
      state := CellState[(iCol + DragCol) div 2,
                         (iRow + DragRow) div 2];
      if white and (state <> csBlack) or
        not white and (state <> csWhite) then Exit;
    end;
    else Exit;
  end;
  Result := True;
end;

procedure TCheckerForm.ExitMIClick(Sender: TObject);
begin
  Close;
end;

procedure TCheckerForm.NewMIClick(Sender: TObject);
begin
  InitBoard;
end;

procedure TCheckerForm.ConnectMIClick(Sender: TObject);
begin
  if ClientSocket.Active then ClientSocket.Active := False;
  if InputQuery('Verbindung mit Gegenspieler', 'PC-Name:', Servername) then
    if Length(Servername) > 0 then
    with ClientSocket do
    begin
      Host := Servername;
      Active := True;
      Connected := True;
    end;
end;

procedure TCheckerForm.Move(FromCell, ToCell: TPoint);
begin
  with FromCell do begin
    DragCol := x;
    DragRow := y;
  end;
  if BeginMove(FromCell.x, FromCell.y) and
    CanMove(ToCell.x, ToCell.y) then
  EndMove(ToCell.x, ToCell.y);
end;

procedure TCheckerForm.SocketRead(Sender: TObject;
  Socket: TCustomWinSocket);
var
  buf: TMoveRecord;
begin
  Socket.ReceiveBuf(buf, SizeOf(buf));
  Move(buf.fromCell, buf.toCell);
  // Der folgende Ausdruck ist wahr, wenn beide
  // Boolean-Variablen den Wert True oder beide
  // den Wert False haben
  ShallMove := not (IsServer xor WhiteMoves);
  UpdateCaption;
end;

procedure TCheckerForm.ServerSocketClientConnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  Connected := True;
  IsServer := True;
  ShallMove := True;
  UpdateCaption;
end;

procedure TCheckerForm.ServerSocketClientDisconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  Connected := False;
end;

procedure TCheckerForm.ClientSocketDisconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  Connected := False;
end;

procedure TCheckerForm.ClientSocketConnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  Connected := True;
  IsServer := False;
  ShallMove := False;
  UpdateCaption;
end;

procedure TCheckerForm.UpdateCaption;
var
  scolor, smessage: string;
begin
  if not Connected then begin
    scolor := ''; smessage := 'warte auf Verbindung...';
  end
  else begin
    if IsServer then scolor := ' (Wei)'
    else scolor := ' (Schwarz)';
    if ShallMove then smessage := 'bitte ziehen!'
    else smessage := 'warte auf Zug der Gegenseite...';
  end;
  Caption := Format('Dame%s - %s', [scolor, smessage]);
end;

end.
