// Beispiel zu Kapitel 6 in "Go To Delphi 5", A.Bohne / G.Lang
//  Addison-Wesley-Longman 1999

// Damespiel (DrawGrid-Beispiel), Version 5

unit CheckerUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, Menus;

const
  NCells = 8;

type
  TCellState = (csNone, csEmpty, csBlack, csWhite);
  TCheckerForm = class(TForm)
    CheckerDrawGrid: TDrawGrid;
    MainMenu1: TMainMenu;
    Spiel1: TMenuItem;
    NewMI: TMenuItem;
    N1: TMenuItem;
    ExitMI: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure CheckerDrawGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormDestroy(Sender: TObject);
    procedure CheckerDrawGridDragOver(Sender, Source: TObject; X,
      Y: Integer; State: TDragState; var Accept: Boolean);
    procedure CheckerDrawGridDragDrop(Sender, Source: TObject; X,
      Y: Integer);
    procedure CheckerDrawGridMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure CheckerDrawGridEndDrag(Sender, Target: TObject; X,
      Y: Integer);
    procedure ExitMIClick(Sender: TObject);
    procedure NewMIClick(Sender: TObject);
  private
    { Private-Deklarationen }
    CellWidth: Integer;
    CellOffset: Integer;
    CellState: array[0..NCells - 1, 0..NCells - 1] of TCellState;
    Palette: HPalette;
    DragRow, DragCol: Integer;
    LastRow, LastCol: Integer;
    WhiteMoves: Boolean;
    IsDragging: Boolean;
    Playing: Boolean;
    Started: Boolean;
    Eaten: Boolean;
    NrPieces: array[False..True] of Integer;
    procedure DrawChecker(Rect: TRect; white: Boolean);
    procedure InitPalette;
    procedure InitBoard;
    function CanMove(iCol, iRow: Integer): Boolean;
  public
    { Public-Deklarationen }
  end;

var
  CheckerForm: TCheckerForm;

implementation

{$R *.DFM}
{$R CURSOR.RES}

const
  NColors = 16;
  crChecker = 1;

{$R-}

procedure TCheckerForm.InitPalette;
// Graustufenpalette erzeugen
var
  i: Integer;
  logPalette: PLogPalette;
begin
  // Speicher fr Record anfordern
  GetMem(logPalette, SizeOf(TLogPalette) +
                     SizeOf(TPaletteEntry) * (NColors - 1));
  with logPalette^ do
  begin
    palVersion := $300;         // Versionsnummer
    palNumEntries := NColors;   // Anzahl Farben
    // Farben fr einzelne Eintrge
    for i := 0 to NColors - 1 do with palPalEntry[i] do
    begin
      peRed := i shl 4;
      peGreen := peRed;
      peBlue := peRed;
      peFlags := 0;
    end;
  end;
  // logische Palette erstellen
  Palette := CreatePalette(LogPalette^);
  FreeMem(logPalette, SizeOf(TLogPalette) +
                      SizeOf(TPaletteEntry) * (NColors - 1));
end;

{$R+}

procedure TCheckerForm.InitBoard;
// Initialisierung der Felder zu Spielbeginn
var
 row, col: Integer;
begin
  if Playing and Started then
  begin
    if MessageDlg('Spiel abbrechen?', mtConfirmation,
                   [mbYes, mbNo], 0) <> mrYes then Exit;
    Playing := False;
  end;
  for row := 0 to NCells - 1 do
    for col := 0 to NCells - 1 do
      if Odd(row + col) then CellState[col, row] := csEmpty
      else CellState[col, row] := csNone;
  for row := 0 to 2 do
    for col := 0 to NCells - 1 do
      if Odd(row + col) then CellState[col, row] := csWhite;
  for row := NCells - 3 to NCells - 1 do
    for col := 0 to NCells - 1 do
      if Odd(row + col) then CellState[col, row] := csBlack;
  CheckerDrawGrid.Invalidate;
  Started := False;
  WhiteMoves := True;
  Eaten := False;
  IsDragging := False;
  NrPieces[False] := 12;
  NrPieces[True] := 12;
end;

procedure TCheckerForm.FormCreate(Sender: TObject);
begin
  // Brettgre einstellen
  with CheckerDrawGrid do
  begin
    DefaultColWidth := Screen.Height div 16;
    DefaultRowHeight := DefaultColWidth;
    Left := 0; Top := 0;
    Width := ColCount * DefaultColWidth + GetSystemMetrics(SM_CXEDGE) * 2;
    Height := RowCount * DefaultRowHeight + GetSystemMetrics(SM_CYEDGE) * 2;
  end;
  ClientWidth := CheckerDrawGrid.Width;
  ClientHeight := CheckerDrawGrid.Height;
  CellWidth := CheckerDrawGrid.DefaultColWidth;
  CellOffset := CellWidth div 16;
  // Graustufenpalette erzeugen
  InitPalette;
  Playing := True;
  Started := False;
  InitBoard;
  Screen.Cursors[crChecker] := LoadCursor(HInstance, 'DRAGCURSOR');
  CheckerDrawGrid.DragCursor := crChecker;
end;

procedure TCheckerForm.FormDestroy(Sender: TObject);
begin
  DeleteObject(Palette);
end;

procedure TCheckerForm.CheckerDrawGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
// Damestein malen
begin
  with CheckerDrawGrid.Canvas do
  begin
    // Hintergrund fllen
    if Odd(ARow + ACol) then Brush.Color := clMaroon
    else Brush.Color := clYellow;
    FillRect(Rect);
    // Stein malen
    if CellState[ACol, ARow] in [csBlack, csWhite] then
    begin
      InflateRect(rect, -CellOffset, -CellOffset);
      DrawChecker(Rect, CellState[ACol, ARow] = csWhite);
    end;
  end;
end;

procedure TCheckerForm.DrawChecker(Rect: TRect; white: Boolean);
const
  NCircles = 18;
  Colors: array [0..NCircles - 1] of TColor =
    (0, 6, 5, 4, 3, 2, 1, 0, 4, 7, 4, 0, 2, 4, 6, 2, 1, 0);
var
  i: Integer;
  w: Integer;
  wrect: Integer;
  oldPalette: HPalette;
begin
  wrect := Rect.Right - Rect.Left;
  with CheckerDrawGrid.Canvas do
  begin
    // Graustufenpalette selektieren
    oldPalette := SelectPalette(Handle, Palette, False);
    RealizePalette(Handle);
    // konzentrische Kreise mit verschiedenen Grauwerten
    // bereinander malen
    for i := 0 to NCircles - 1 do
    begin
      w := Round(wrect * (1.0 - 0.025 * i));
      if white then Brush.Color := PaletteIndex(Colors[i] + 8)
      else Brush.Color := PaletteIndex(Colors[i]);
      Pen.Color := Brush.Color;
      Ellipse(Rect.Left + w, Rect.Top + w,
              Rect.Right - w, Rect.Bottom - w);
    end;
    // alte Palette selektieren
    SelectPalette(Handle, oldPalette, False);
  end;
end;

procedure TCheckerForm.CheckerDrawGridDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
// entscheiden, ob das Spielfeld unter dem Cursor
// das "Fallenlassen" erlaubt (wird beim Bewegen des Cursors
// ber ein neues Spielfeld aufgerufen)
var
  iCol, iRow: Integer;
begin
  if Source is TDrawGrid then
  begin
    CheckerDrawGrid.MouseToCell(X, Y, iCol, iRow);
    Accept := (iCol >= 0) and (iCol < NCells)
               and (iRow >= 0) and (iRow < NCells)
               and CanMove(iCol, iRow);
  end
  else
    Accept := False;
end;

procedure TCheckerForm.CheckerDrawGridDragDrop(Sender, Source: TObject; X,
  Y: Integer);
// Nach Fallenlassen Spielfeld aktualisieren
var
  iCol, iRow: Integer;
  rect: TRect;
  msg: string;
begin
  CheckerDrawGrid.MouseToCell(X, Y, iCol, iRow);
  // Zielfeld bekommt Status von Ausgangsfeld
  CellState[iCol, iRow] := CellState[DragCol, DragRow];
  LastRow := iRow; LastCol := iCol;
  // Ausgangsfeld ist jetzt leer
  CellState[DragCol, DragRow] := csEmpty;
  with CheckerDrawGrid do
  begin
    // Zielfeld neu zeichnen lassen
    rect := CellRect(iCol, iRow);
    InvalidateRect(Handle, @rect, True);
    // Schlagen behandeln
    if Abs(iRow - DragRow) = 2 then
    begin
      if Eaten and (WhiteMoves <>
          (CellState[iCol, iRow] = csWhite)) then
        WhiteMoves := not WhiteMoves;
      iCol := (iCol + DragCol) div 2;
      iRow := (iRow + DragRow) div 2;
      // Feld mit geschlagener Figur leeren
      CellState[iCol, iRow] := csEmpty;
      rect := CellRect(iCol, iRow);
      InvalidateRect(Handle, @rect, True);
      Eaten := True;
      Dec(NrPieces[not WhiteMoves]);
      // alle Steine einer Farbe geschlagen
      if NrPieces[not WhiteMoves] = 0 then
      begin
        if WhiteMoves then msg := 'Wei'
        else msg := 'Schwarz';
        ShowMessage(msg + ' hat gewonnen!');
        Playing := False;
        Exit;
      end;
    end
    else Eaten := False;
  end;
  WhiteMoves := not WhiteMoves;
  IsDragging := False;
  Started := True;
end;

procedure TCheckerForm.CheckerDrawGridMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
// beginnt Ziehen, wenn erlaubt
var
  rect: TRect;
  state: TCellState;
  white: boolean;
begin
  if Playing and (Button = mbLeft) then with CheckerDrawGrid do
  begin
    MouseToCell(X, Y, DragCol, DragRow);
    state := CellState[DragCol, DragRow];
    white := state = csWhite;
    // Wenn am Zug
    if (WhiteMoves = white) or Eaten and (DragCol = LastCol)
        and (DragRow = LastRow) then begin
      // Ausgangsfeld leeren neu zeichnen...
      CellState[DragCol, DragRow] := csEmpty;
      rect := CellRect(DragCol, DragRow);
      InvalidateRect(Handle, @rect, False);
      Update;
      // ...und wieder auf alten Status zurcksetzen
      // (macht Abbruch einfacher)
      CellState[DragCol, DragRow] := state;
      IsDragging := True;
      BeginDrag(True);
    end;
  end;
end;

procedure TCheckerForm.CheckerDrawGridEndDrag(Sender, Target: TObject; X,
  Y: Integer);
// zeichnet bei Abbruch gezogenen Stein an Ausgangsposition
var
  rect: TRect;
begin
  if IsDragging then with CheckerDrawGrid do
  begin
    rect := CellRect(DragCol, DragRow);
    InvalidateRect(Handle, @rect, False);
    IsDragging := False;
  end;
end;

function TCheckerForm.CanMove(iCol, iRow: Integer): Boolean;
// prft, ob ein Zug erlaubt ist
var
  white: Boolean;
  state: TCellState;
begin
  Result := False;
  // nur auf leere Zellen ziehen lassen
  if CellState[iCol, iRow] <> csEmpty then Exit;
  white := CellState[DragCol, DragRow] = csWhite;
  // Rckwrtsgehen verbieten
  if white and (iRow <= DragRow) or
    not white and (iRow >= DragRow) then Exit;
  case Abs(iRow - DragRow) of
    // einfacher Zug
    1: if (white <> WhiteMoves) or (Abs(iCol - DragCol) <> 1) then Exit;
    // Schlagen
    2:
    begin
      if Abs(iCol - DragCol) <> 2 then Exit;
      state := CellState[(iCol + DragCol) div 2,
                         (iRow + DragRow) div 2];
      if white and (state <> csBlack) or
        not white and (state <> csWhite) then Exit;
    end;
    else Exit;
  end;
  Result := True;
end;

procedure TCheckerForm.ExitMIClick(Sender: TObject);
begin
  Close;
end;

procedure TCheckerForm.NewMIClick(Sender: TObject);
begin
  InitBoard;
end;

end.
